from flask import render_template, redirect, url_for, flash, Blueprint
from sqlalchemy.exc import OperationalError
from celery import states
from celery.utils.log import get_task_logger
from flask_admin.contrib.sqla import ModelView
from flask_login import login_required

from database import app, admin, db
from covid19.blueprints.application.application_services import divi_service
from covid19.blueprints.application.application_workers import celery

from covid19.blueprints.intensivregister.divi_model import DiviRegion, DiviDateReported, DiviCountry, DiviData
from covid19.blueprints.intensivregister.divi_model_import import DiviImport
from covid19.blueprints.application.application_model_transient import ApplicationPage


app_divi = Blueprint('divi', __name__, template_folder='templates', url_prefix='/divi')

admin.add_view(ModelView(DiviImport, db.session, category="DIVI"))
admin.add_view(ModelView(DiviDateReported, db.session, category="DIVI"))
admin.add_view(ModelView(DiviRegion, db.session, category="DIVI"))
admin.add_view(ModelView(DiviCountry, db.session, category="DIVI"))
admin.add_view(ModelView(DiviData, db.session, category="DIVI"))


# ---------------------------------------------------------------------------------------------------------------
#  Url Routes Frontend
# ---------------------------------------------------------------------------------------------------------------


@app_divi.route('/info')
def url_divi_info():
    page_info = ApplicationPage('WHO', "Info")
    return render_template(
        'divi/divi_info.html',
        page_info=page_info)


@app_divi.route('/tasks')
@login_required
def url_divi_tasks():
    page_info = ApplicationPage('WHO', "Tasks")
    return render_template(
        'divi/divi_tasks.html',
        page_info=page_info)


@app_divi.route('/imported/page/<int:page>')
@app_divi.route('/imported')
@login_required
def url_divi_imported(page=1):
    page_info = ApplicationPage('WHO', "Last Import")
    try:
        page_data = DiviImport.get_all_as_page(page)
    except OperationalError:
        flash("No data in the database.")
        page_data = None
    return render_template(
        'divi/divi_imported.html',
        page_data=page_data,
        page_info=page_info)


@app_divi.route('/date_reported/all/page/<int:page>')
@app_divi.route('/date_reported/all')
def url_divi_date_reported_all(page: int = 1):
    page_info = ApplicationPage('WHO', "Date Reported", "All")
    try:
        page_data = DiviDateReported.get_all_as_page(page)
    except OperationalError:
        flash("No regions in the database.")
        page_data = None
    return render_template(
        'divi/date_reported/divi_date_reported_all.html',
        page_data=page_data,
        page_info=page_info)


@app_divi.route('/date_reported/<int:date_reported_id>/page/<int:page>')
@app_divi.route('/date_reported/<int:date_reported_id>')
def url_divi_date_reported(date_reported_id: int, page: int = 1):
    date_reported = DiviDateReported.get_by_id(date_reported_id)
    page_info = ApplicationPage(
        "Date Reported: " + date_reported.date_reported,
        'WHO',
        "data of all reported countries for WHO date reported " + date_reported.date_reported + " "
    )
    try:
        page_data = DiviData.get_data_for_day(date_reported, page)
    except OperationalError:
        flash("No data in the database.")
        page_data = None
    return render_template(
        'divi/date_reported/divi_date_reported_one.html',
        who_date_reported=date_reported,
        page_data=page_data,
        page_info=page_info)


@app_divi.route('/date_reported/<int:date_reported_id>/cases_new/page/<int:page>')
@app_divi.route('/date_reported/<int:date_reported_id>/cases_new')
def url_divi_date_reported_cases_new(date_reported_id: int, page: int = 1):
    date_reported = DiviDateReported.get_by_id(date_reported_id)
    page_info = ApplicationPage(
        "Date Reported: " + date_reported.date_reported,
        'WHO',
        "data of all reported countries for WHO date reported " + date_reported.date_reported + " "
    )
    try:
        page_data = DiviData.get_data_for_day_order_by_cases_new(date_reported, page)
    except OperationalError:
        flash("No data in the database.")
        page_data = None
    return render_template(
        'divi/date_reported/divi_date_reported_one_cases_new.html',
        who_date_reported=date_reported,
        page_data=page_data,
        page_info=page_info)


@app_divi.route('/date_reported/<int:date_reported_id>/cases_cumulative/page/<int:page>')
@app_divi.route('/date_reported/<int:date_reported_id>/cases_cumulative')
def url_divi_date_reported_cases_cumulative(date_reported_id: int, page: int = 1):
    date_reported = DiviDateReported.get_by_id(date_reported_id)
    page_info = ApplicationPage(
        "Date Reported: " + date_reported.date_reported,
        'WHO',
        "data of all reported countries for WHO date reported " + date_reported.date_reported + " "
    )
    try:
        page_data = DiviData.get_data_for_day_order_by_cases_cumulative(date_reported, page)
    except OperationalError:
        flash("No data in the database.")
        page_data = None
    return render_template(
        'divi/date_reported/divi_date_reported_one_cases_cumulative.html',
        who_date_reported=date_reported,
        page_data=page_data,
        page_info=page_info)


@app_divi.route('/date_reported/<int:date_reported_id>/deaths_new/page/<int:page>')
@app_divi.route('/date_reported/<int:date_reported_id>/deaths_new')
def url_divi_date_reported_deaths_new(date_reported_id: int, page: int = 1):
    date_reported = DiviDateReported.get_by_id(date_reported_id)
    page_info = ApplicationPage(
        "Date Reported: " + date_reported.date_reported,
        'WHO',
        "data of all reported countries for WHO date reported " + date_reported.date_reported + " "
    )
    try:
        page_data = DiviData.get_data_for_day_order_by_deaths_new(date_reported, page)
    except OperationalError:
        flash("No data in the database.")
        page_data = None
    return render_template(
        'divi/date_reported/divi_date_reported_one_deaths_new.html',
        who_date_reported=date_reported,
        page_data=page_data,
        page_info=page_info)


@app_divi.route('/date_reported/<int:date_reported_id>/deaths_cumulative/page/<int:page>')
@app_divi.route('/date_reported/<int:date_reported_id>/deaths_cumulative')
def url_divi_date_reported_deaths_cumulative(date_reported_id: int, page: int = 1):
    date_reported = DiviDateReported.get_by_id(date_reported_id)
    page_info = ApplicationPage(
        "Date Reported: " + date_reported.date_reported,
        'WHO',
        "data of all reported countries for WHO date reported " + date_reported.date_reported + " "
    )
    try:
        page_data = DiviData.get_data_for_day_order_by_deaths_cumulative(date_reported, page)
    except OperationalError:
        flash("No data in the database.")
        page_data = None
    return render_template(
        'divi/date_reported/divi_date_reported_one_deaths_cumulative.html',
        who_date_reported=date_reported,
        page_data=page_data,
        page_info=page_info)


@app_divi.route('/region/all/page/<int:page>')
@app_divi.route('/region/all')
def url_divi_region_all(page: int = 1):
    page_info = ApplicationPage('WHO', "Region", "All")
    try:
        page_data = DiviRegion.get_all_as_page(page)
    except OperationalError:
        flash("No regions in the database.")
        page_data = None
    return render_template(
        'divi/region/divi_region_all.html',
        page_data=page_data,
        page_info=page_info)


@app_divi.route('/region/<int:region_id>/page/<int:page>')
@app_divi.route('/region/<int:region_id>')
def url_divi_region(region_id: int, page: int = 1):
    who_region = None
    page_info = ApplicationPage("Countries", "WHO Region")
    try:
        who_region = DiviRegion.get_by_id(region_id)
        page_data = DiviCountry.get_who_countries_for_region(who_region, page)
        page_info.title = who_region.region
        page_info.subtitle = "WHO Region"
        page_info.subtitle_info = "Countries of WHO Region " + who_region.region
    except OperationalError:
        flash("No countries of that region in the database.")
        page_data = None
    return render_template(
        'divi/region/divi_region_one.html',
        who_region=who_region,
        page_data=page_data,
        page_info=page_info)


@app_divi.route('/country/all/page/<int:page>')
@app_divi.route('/country/all')
def url_divi_country_all(page: int = 1):
    page_info = ApplicationPage('WHO', "Countries", "All")
    try:
        page_data = DiviCountry.get_all_as_page(page)
    except OperationalError:
        flash("No regions in the database.")
        page_data = None
    return render_template(
        'divi/country/divi_country_all.html',
        page_data=page_data,
        page_info=page_info)


@app_divi.route('/country/<int:country_id>/page/<int:page>')
@app_divi.route('/country/<int:country_id>')
def url_divi_country(country_id: int, page: int = 1):
    who_country = DiviCountry.get_by_id(country_id)
    page_data = DiviData.get_data_for_country(who_country, page)
    page_info = ApplicationPage(who_country.country,
           "Country "+who_country.country_code,
           "Data per Day in Country "+who_country.country+" of WHO Region "+who_country.region.region)
    return render_template(
        'divi/country/divi_country_one.html',
        who_country=who_country,
        page_data=page_data,
        page_info=page_info)


@app_divi.route('/country/<int:country_id>/cases_new/page/<int:page>')
@app_divi.route('/country/<int:country_id>/cases_new')
def url_divi_country_cases_new(country_id: int, page: int = 1):
    who_country = DiviCountry.get_by_id(country_id)
    page_data = DiviData.get_data_for_country_order_by_cases_new(who_country, page)
    page_info = ApplicationPage(who_country.country,
           "Country "+who_country.country_code,
           "Data per Day in Country "+who_country.country+" of WHO Region "+who_country.region.region)
    return render_template(
        'divi/country/divi_country_one_cases_new.html',
        who_country=who_country,
        page_data=page_data,
        page_info=page_info)


@app_divi.route('/country/<int:country_id>/cases_cumulative/page/<int:page>')
@app_divi.route('/country/<int:country_id>/cases_cumulative')
def url_divi_country_cases_cumulative(country_id: int, page: int = 1):
    who_country = DiviCountry.get_by_id(country_id)
    page_data = DiviData.get_data_for_country_order_by_cases_cumulative(who_country, page)
    page_info = ApplicationPage(who_country.country,
           "Country "+who_country.country_code,
           "Data per Day in Country "+who_country.country+" of WHO Region "+who_country.region.region)
    return render_template(
        'divi/country/divi_country_one_cases_cumulative.html',
        who_country=who_country,
        page_data=page_data,
        page_info=page_info)


@app_divi.route('/country/<int:country_id>/deaths_new/page/<int:page>')
@app_divi.route('/country/<int:country_id>/deaths_new')
def url_divi_country_deaths_new(country_id: int, page: int = 1):
    who_country = DiviCountry.get_by_id(country_id)
    page_data = DiviData.get_data_for_country_order_by_deaths_new(who_country, page)
    page_info = ApplicationPage(who_country.country,
           "Country "+who_country.country_code,
           "Data per Day in Country "+who_country.country+" of WHO Region "+who_country.region.region)
    return render_template(
        'divi/country/divi_country_one_deaths_new.html',
        who_country=who_country,
        page_data=page_data,
        page_info=page_info)


@app_divi.route('/country/<int:country_id>/deaths_cumulative/page/<int:page>')
@app_divi.route('/country/<int:country_id>/deaths_cumulative')
def url_divi_country_deaths_cumulative(country_id: int, page: int = 1):
    who_country = DiviCountry.get_by_id(country_id)
    page_data = DiviData.get_data_for_country_order_by_deaths_cumulative(who_country, page)
    page_info = ApplicationPage(who_country.country,
           "Country "+who_country.country_code,
           "Data per Day in Country "+who_country.country+" of WHO Region "+who_country.region.region)
    return render_template(
        'divi/country/divi_country_one_deaths_cumulative.html',
        who_country=who_country,
        page_data=page_data,
        page_info=page_info)


@app_divi.route('/germany/page/<int:page>')
@app_divi.route('/germany')
def url_divi_germany(page: int = 1):
    page_info = ApplicationPage('WHO', "Germany")
    who_country_germany = DiviCountry.get_germany()
    if who_country_germany is None:
        flash('country: Germany not found in Database', category='error')
        return redirect(url_for('who.url_divi_tasks'))
    page_data = DiviData.get_data_for_country(who_country_germany, page)
    return render_template(
        'divi/country/divi_country_germany.html',
        who_country=who_country_germany,
        page_data=page_data,
        page_info=page_info)


@app_divi.route('/mytest')
def url_divi_mytest():
    flash("url_divi_mytest - START: DiviImport.countries()")
    app.logger.info("url_divi_mytest - START: DiviImport.countries()")
    i = 0
    for c in DiviImport.countries():
        i += 1
        line = " | " + str(i) + " | " + c.countries.country_code + " | " + c.countries.country + " | " + c.countries.who_region + " | "
        app.logger.info(line)
    flash("url_divi_mytest - DONE: DiviImport.countries()")
    app.logger.info("url_divi_mytest - DONE: DiviImport.countries()")
    flash("url_divi_mytest - START: DiviImport.get_new_dates_as_array()")
    app.logger.info("url_divi_mytest - START: DiviImport.get_new_dates_as_array()")
    i = 0
    for date_reported in DiviImport.get_new_dates_as_array():
        i += 1
        line = " | " + str(i) + " | " + date_reported + " | "
        app.logger.info(line)
    joungest_datum = DiviDateReported.get_joungest_datum()
    app.logger.info(joungest_datum)
    i = 0
    for who_data in DiviData.get_data_for_one_day(joungest_datum):
        i += 1
        line = " | " + str(i) + " | " + str(who_data.date_reported) + " | " + who_data.country.country + " | "
        app.logger.info(line)
    flash("url_divi_mytest - DONE: DiviImport.get_new_dates_as_array()")
    app.logger.info("url_divi_mytest - DONE: DiviImport.get_new_dates_as_array()")
    return redirect(url_for('who.url_divi_tasks'))

# ----------------------------------------------------------------------------------------------------------------
#  Celery TASKS
# ----------------------------------------------------------------------------------------------------------------


@celery.task(bind=True)
def task_divi_download_only(self):
    logger = get_task_logger(__name__)
    self.update_state(state=states.STARTED)
    logger.info("------------------------------------------------------------")
    logger.info(" Received: task_divi_download_only [OK] ")
    logger.info("------------------------------------------------------------")
    divi_service.run_download_only()
    self.update_state(state=states.SUCCESS)
    result = "OK (task_divi_download_only)"
    return result


@celery.task(bind=True)
def task_divi_import_only(self):
    logger = get_task_logger(__name__)
    self.update_state(state=states.STARTED)
    logger.info("------------------------------------------------------------")
    logger.info(" Received: task_divi_import_only [OK] ")
    logger.info("------------------------------------------------------------")
    divi_service.run_import_only()
    self.update_state(state=states.SUCCESS)
    result = "OK (task_divi_import_only)"
    return result


@celery.task(bind=True)
def task_divi_update_dimension_tables_only(self):
    logger = get_task_logger(__name__)
    self.update_state(state=states.STARTED)
    logger.info("------------------------------------------------------------")
    logger.info(" Received: task_divi_update_dimension_tables_only [OK] ")
    logger.info("------------------------------------------------------------")
    divi_service.run_update_dimension_tables_only()
    self.update_state(state=states.SUCCESS)
    result = "OK (task_divi_update_dimension_tables_only)"
    return result


@celery.task(bind=True)
def task_divi_update_fact_table_incremental_only(self):
    logger = get_task_logger(__name__)
    self.update_state(state=states.STARTED)
    logger.info("------------------------------------------------------------")
    logger.info(" Received: task_divi_update_fact_table_incremental_only [OK] ")
    logger.info("------------------------------------------------------------")
    divi_service.run_update_fact_table_incremental_only()
    self.update_state(state=states.SUCCESS)
    result = "OK (task_divi_update_dimension_tables_only)"
    return result


@celery.task(bind=True)
def task_divi_update_fact_table_initial_only(self):
    logger = get_task_logger(__name__)
    self.update_state(state=states.STARTED)
    logger.info("------------------------------------------------------------")
    logger.info(" Received: task_divi_update_fact_table_initial_only [OK] ")
    logger.info("------------------------------------------------------------")
    divi_service.run_update_fact_table_initial_only()
    self.update_state(state=states.SUCCESS)
    result = "OK (task_divi_update_fact_table_initial_only)"
    return result


@celery.task(bind=True)
def task_divi_update_star_schema_incremental(self):
    logger = get_task_logger(__name__)
    self.update_state(state=states.STARTED)
    logger.info("------------------------------------------------------------")
    logger.info(" Received: task_divi_update_star_schema_incremental [OK] ")
    logger.info("------------------------------------------------------------")
    divi_service.run_update_star_schema_incremental()
    self.update_state(state=states.SUCCESS)
    result = "OK (task_divi_update_star_schema_incremental)"
    return result


@celery.task(bind=True)
def task_divi_update_star_schema_initial(self):
    logger = get_task_logger(__name__)
    self.update_state(state=states.STARTED)
    logger.info("------------------------------------------------------------")
    logger.info(" Received: task_divi_update_star_schema_initial [OK] ")
    logger.info("------------------------------------------------------------")
    divi_service.run_update_star_schema_initial()
    self.update_state(state=states.SUCCESS)
    result = "OK (task_divi_update_star_schema_incremental)"
    return result

# ----------------------------------------------------------------------------------------------------------------
#  URL Routes for Celery TASKS
# ----------------------------------------------------------------------------------------------------------------


@app_divi.route('/task/download/only')
@login_required
def url_task_divi_download_only():
    app.logger.info("url_divi_task_download_only [start]")
    divi_service.run_download_only()
    flash("divi_service.run_download_only ok")
    app.logger.info("url_divi_task_download_only [done]")
    return redirect(url_for('who.url_divi_tasks'))


@app_divi.route('/task/import/only')
@login_required
def url_task_divi_import_only():
    app.logger.info("url_divi_update_run [start]")
    task_divi_import_only.apply_async()
    flash("divi_service.run_update started")
    flash(message="long running background task started", category="warning")
    app.logger.info("url_divi_update_run [done]")
    return redirect(url_for('who.url_divi_tasks'))


@app_divi.route('/task/update/dimension-tables/only')
@login_required
def url_task_divi_update_dimension_tables_only():
    app.logger.info("url_task_divi_update_dimension_tables_only [start]")
    task_divi_update_dimension_tables_only.apply_async()
    flash("task_divi_update_dimension_tables_only started")
    flash(message="long running background task started", category="warning")
    app.logger.info("url_task_divi_update_dimension_tables_only [done]")
    return redirect(url_for('who.url_divi_tasks'))


@app_divi.route('/task/update/fact-table/incremental/only')
@login_required
def url_task_divi_update_fact_table_incremental_only():
    app.logger.info("url_task_divi_update_fact_table_incremental_only [start]")
    task_divi_update_fact_table_incremental_only.apply_async()
    flash("task_divi_update_fact_table_incremental_only started")
    flash(message="long running background task started", category="warning")
    app.logger.info("url_task_divi_update_fact_table_incremental_only [done]")
    return redirect(url_for('who.url_divi_tasks'))


@app_divi.route('/task/update/fact-table/initial/only')
@login_required
def url_task_divi_update_fact_table_initial_only():
    app.logger.info("url_task_divi_update_fact_table_initial_only [start]")
    task_divi_update_fact_table_initial_only.apply_async()
    flash("task_divi_update_fact_table_initial_only started")
    flash(message="long running background task started", category="warning")
    app.logger.info("url_divi_task_update_full [done]")
    return redirect(url_for('who.url_divi_tasks'))


@app_divi.route('/task/update/star_schema/initial')
@login_required
def url_task_divi_update_star_schema_initial():
    app.logger.info("url_divi_task_update_full [start]")
    divi_service.run_download_only()
    flash("divi_service.service_download.download_file ok")
    task_divi_update_star_schema_initial.apply_async()
    flash("task_divi_update_star_schema_initial started")
    flash(message="long running background task started", category="warning")
    app.logger.info("url_divi_task_update_full [done]")
    return redirect(url_for('who.url_divi_tasks'))


@app_divi.route('/task/update/star_schema/incremental')
@login_required
def url_task_divi_update_star_schema_incremental():
    app.logger.info("url_task_divi_update_star_schema_incremental [start]")
    divi_service.run_download_only()
    flash("divi_service.service_download.download_file ok")
    task_divi_update_star_schema_incremental.apply_async()
    flash("task_divi_run_update_full started")
    flash(message="long running background task started", category="warning")
    app.logger.info("url_task_divi_update_star_schema_incremental [done]")
    return redirect(url_for('who.url_divi_tasks'))
